@@ -1,7 +1,7 @@
 #!python
 ##############################################################################
 #
-# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
+# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -16,18 +16,17 @@
 
 Usage: python zdctl.py [-C URL] [-S schema.xml] [-h] [-p PROGRAM]
        [zdrun-options] [action [arguments]]
- 
+
 Options:
--C/--configure URL -- configuration file or URL
--S/--schema XML Schema -- XML schema for configuration file
--h/--help -- print usage message and exit
 -b/--backoff-limit SECONDS -- set backoff limit to SECONDS (default 10)
+-C/--configure URL -- configuration file or URL
 -d/--daemon -- run as a proper daemon; fork a subprocess, close files etc.
 -f/--forever -- run forever (by default, exit when backoff limit is exceeded)
 -h/--help -- print this usage message and exit
--i/--interactive -- start an interactive shell after executing commands
 -l/--logfile -- log file to be read by logtail command
 -p/--program PROGRAM -- the program to run
+-S/--schema XML Schema -- XML schema for configuration file
+-T/--start-timeout SECONDS -- Start timeout when a test program is used
 -s/--socket-name SOCKET -- Unix socket name for client (default "zdsock")
 -u/--user USER -- run as this user (or numeric uid)
 -m/--umask UMASK -- use this umask for daemon subprocess (default is 022)
@@ -35,10 +34,7 @@ Options:
 -z/--directory DIRECTORY -- directory to chdir to when using -d (default off)
 action [arguments] -- see below
 
-Actions are commands like "start", "stop" and "status".  If -i is
-specified or no action is specified on the command line, a "shell"
-interpreting actions typed interactively is started (unless the
-configuration option default_to_interactive is set to false).  Use the
+Actions are commands like "start", "stop" and "status".  Use the
 action "help" to find out about available actions.
 """
 
@@ -76,9 +72,9 @@ def string_list(arg):
 
 
 class ZDCtlOptions(RunnerOptions):
-    
+
     __doc__ = __doc__
-    
+
     positional_args_allowed = True
 
     def __init__(self):
@@ -86,13 +82,12 @@ class ZDCtlOptions(RunnerOptions):
         self.add("schemafile", short="S:", long="schema=",
                  default="schema.xml",
                  handler=self.set_schemafile)
-        self.add("interactive", None, "i", "interactive", flag=1)
-        self.add("default_to_interactive", "runner.default_to_interactive",
-                 default=1)
         self.add("program", "runner.program", "p:", "program=",
                  handler=string_list,
                  required="no program specified; use -p or -C")
         self.add("logfile", "runner.logfile", "l:", "logfile=")
+        self.add("start_timeout", "runner.start_timeout",
+                 "T:", "start-timeout=", int, default=300)
         self.add("python", "runner.python")
         self.add("zdrun", "runner.zdrun")
         programname = os.path.basename(sys.argv[0])
@@ -102,14 +97,12 @@ class ZDCtlOptions(RunnerOptions):
         self.add("prompt", "runner.prompt", default=(programname + ">"))
 
     def realize(self, *args, **kwds):
-        
+
         RunnerOptions.realize(self, *args, **kwds)
 
         # Maybe the config file requires -i or positional args
-        if not self.args and not self.interactive:
-            if not self.default_to_interactive:
-                self.usage("either -i or an action argument is required")
-            self.interactive = 1
+        if not self.args:
+            self.usage("an action argument is required")
 
         # Where's python?
         if not self.python:
@@ -173,14 +166,42 @@ class ZDCmd(cmd.Cmd):
             os.chown(directory, self.options.uid, self.options.gid)
 
     def set_uid(self):
-        if self.options.uid is None:
+        user = self.options.user
+        if user is None:
             return
-        uid = os.geteuid()
-        if uid != 0 and uid != self.options.uid:
-            self.options.usage("only root can use -u USER to change users")
-        os.setgid(self.options.gid)
-        os.setgroups(self.options.groups)
-        os.setuid(self.options.uid)
+
+        import pwd
+        try:
+            uid = int(user)
+        except ValueError:
+            try:
+                pwrec = pwd.getpwnam(user)
+            except KeyError:
+                self.options.usage("username %r not found" % user)
+            uid = pwrec.pw_uid
+        else:
+            try:
+                pwrec = pwd.getpwuid(uid)
+            except KeyError:
+                self.options.usage("uid %r not found" % user)
+
+        # See if we're already that user:
+        euid = os.geteuid()
+        if euid != 0:
+            if euid != uid:
+                self.options.usage("only root can use -u USER to change users")
+            return
+
+        # OK, we have to set user and groups:
+        os.setgid(pwrec.pw_gid)
+
+        import grp
+        user = pwrec.pw_name
+        os.setgroups(
+            sorted(g.gr_gid for g in grp.getgrall() # sort for tests
+                   if user in g.gr_mem)
+            )
+        os.setuid(uid)
 
     def emptyline(self):
         # We don't want a blank line to repeat the last command.
@@ -208,6 +229,7 @@ class ZDCmd(cmd.Cmd):
         except socket.error, msg:
             return None
 
+    zd_testing = 0
     def get_status(self):
         self.zd_up = 0
         self.zd_pid = 0
@@ -221,6 +243,12 @@ class ZDCmd(cmd.Cmd):
         self.zd_up = 1
         self.zd_pid = int(m.group(1))
         self.zd_status = resp
+        m = re.search("(?m)^testing=(\d+)$", resp)
+        if m:
+            self.zd_testing = int(m.group(1))
+        else:
+            self.zd_testing = 0
+
         return resp
 
     def awhile(self, cond, msg):
@@ -230,14 +258,14 @@ class ZDCmd(cmd.Cmd):
             if self.get_status():
                 was_running = True
 
-            while not cond():
+            while not cond(n):
                 sys.stdout.write(". ")
                 sys.stdout.flush()
                 time.sleep(1)
                 n += 1
                 if self.get_status():
                     was_running = True
-                elif (was_running or n > 10) and not cond():
+                elif (was_running or n > 10) and not cond(n):
                     print "\ndaemon manager not running"
                     return
 
@@ -250,12 +278,11 @@ class ZDCmd(cmd.Cmd):
         print "help          -- Print a list of available actions."
         print "help <action> -- Print help for <action>."
 
-    def do_EOF(self, arg):
-        print
-        return 1
-
-    def help_EOF(self):
-        print "To quit, type ^D or use the quit command."
+    def _start_cond(self, n):
+        if (n > self.options.start_timeout):
+            print '\nProgram took too long to start'
+            sys.exit(1)
+        return self.zd_pid and not self.zd_testing
 
     def do_start(self, arg):
         self.get_status()
@@ -291,8 +318,9 @@ class ZDCmd(cmd.Cmd):
             print "daemon process already running; pid=%d" % self.zd_pid
             return
         if self.options.daemon:
-            self.awhile(lambda: self.zd_pid,
-                        "daemon process started, pid=%(zd_pid)d")
+            self.awhile(self._start_cond,
+                        "daemon process started, pid=%(zd_pid)d",
+                        )
 
     def _get_override(self, opt, name, svalue=None, flag=0):
         value = getattr(self.options, name)
@@ -333,7 +361,7 @@ class ZDCmd(cmd.Cmd):
             print "daemon process not running"
         else:
             self.send_action("stop")
-            self.awhile(lambda: not self.zd_pid, "daemon process stopped")
+            self.awhile(lambda n: not self.zd_pid, "daemon process stopped")
 
     def do_reopen_transcript(self, arg):
         if not self.zd_up:
@@ -354,7 +382,7 @@ class ZDCmd(cmd.Cmd):
             self.do_start(arg)
         else:
             self.send_action("restart")
-            self.awhile(lambda: self.zd_pid not in (0, pid),
+            self.awhile(lambda n: (self.zd_pid != pid) and self._start_cond(n),
                         "daemon process restarted, pid=%(zd_pid)d")
 
     def help_restart(self):
@@ -386,7 +414,7 @@ class ZDCmd(cmd.Cmd):
         print "              The default signal is SIGTERM."
 
     def do_wait(self, arg):
-        self.awhile(lambda: not self.zd_pid, "daemon process stopped")
+        self.awhile(lambda n: not self.zd_pid, "daemon process stopped")
         self.do_status()
 
     def help_wait(self):
@@ -425,9 +453,6 @@ class ZDCmd(cmd.Cmd):
         print "zdctl/zdrun options:"
         print "schemafile:  ", repr(self.options.schemafile)
         print "configfile:  ", repr(self.options.configfile)
-        print "interactive: ", repr(self.options.interactive)
-        print "default_to_interactive:",
-        print                  repr(self.options.default_to_interactive)
         print "zdrun:       ", repr(self.options.zdrun)
         print "python:      ", repr(self.options.python)
         print "program:     ", repr(self.options.program)
@@ -445,7 +470,6 @@ class ZDCmd(cmd.Cmd):
         print "umask:       ", oct(umask)
         print "directory:   ", repr(self.options.directory)
         print "logfile:     ", repr(self.options.logfile)
-        print "hang_around: ", repr(self.options.hang_around)
 
     def show_python(self):
         print "Python info:"
@@ -469,17 +493,6 @@ class ZDCmd(cmd.Cmd):
         print "show python -- show Python version and details"
         print "show all -- show all of the above"
 
-    def complete_show(self, text, *ignored):
-        options = ["options", "python", "all"]
-        return [x for x in options if x.startswith(text)]
-
-    def do_logreopen(self, arg):
-        self.do_kill(str(signal.SIGUSR2))
-
-    def help_logreopen(self):
-        print "logreopen -- Send a SIGUSR2 signal to the daemon process."
-        print "             This is designed to reopen the log file."
-
     def do_logtail(self, arg):
         if not arg:
             arg = self.options.logfile
@@ -501,46 +514,6 @@ class ZDCmd(cmd.Cmd):
         print "                     A default file may exist."
         print "                     Hit ^C to exit this mode."
 
-    def do_shell(self, arg):
-        if not arg:
-            arg = os.getenv("SHELL") or "/bin/sh"
-        try:
-            os.system(arg)
-        except KeyboardInterrupt:
-            print
-
-    def help_shell(self):
-        print "shell [command] -- Execute a shell command."
-        print "                   Without a command, start an interactive sh."
-        print "An alias for this command is ! [command]"
-
-    def do_reload(self, arg):
-        if arg:
-            args = arg.split()
-            if self.options.configfile:
-                args = ["-C", self.options.configfile] + args
-        else:
-            args = None
-        options = ZDCtlOptions()
-        options.positional_args_allowed = 0
-        try:
-            options.realize(args)
-        except SystemExit:
-            print "Configuration not reloaded"
-        else:
-            self.options = options
-            if self.options.configfile:
-                print "Configuration reloaded from", self.options.configfile
-            else:
-                print "Configuration reloaded without a config file"
-
-    def help_reload(self):
-        print "reload [options] -- Reload the configuration."
-        print "    Without options, this reparses the command line."
-        print "    With options, this substitutes 'options' for the"
-        print "    command line, except that if no -C option is given,"
-        print "    the last configuration file is used."
-
     def do_foreground(self, arg):
         self.get_status()
         pid = self.zd_pid
@@ -566,23 +539,6 @@ class ZDCmd(cmd.Cmd):
     def help_fg(self):
         self.help_foreground()
 
-    def do_quit(self, arg):
-        self.get_status()
-        if not self.zd_up:
-            print "daemon manager not running"
-        elif not self.zd_pid:
-            print "daemon process not running; stopping daemon manager"
-            self.send_action("exit")
-            self.awhile(lambda: not self.zd_up, "daemon manager stopped")
-        else:
-            print "daemon process and daemon manager still running"
-        return 1
-
-    def help_quit(self):
-        print "quit -- Exit the zdctl shell."
-        print "        If the daemon process is not running,"
-        print "        stop the daemon manager."
-
 
 class TailHelper:
 
@@ -654,16 +610,7 @@ def main(args=None, options=None, cmdclass=ZDCmd):
         options = ZDCtlOptions()
     options.realize(args)
     c = cmdclass(options)
-    if options.args:
-        c.onecmd(" ".join(options.args))
-    if options.interactive:
-        try:
-            import readline
-        except ImportError:
-            pass
-        print "program:", " ".join(options.program)
-        c.do_status()
-        c.cmdloop()
+    c.onecmd(" ".join(options.args))
 
 if __name__ == "__main__":
     main()
